/*
 * delstr.c
 *
 * Part of the strings db toolkit.
 *
 * Delstr marks a record, indicated by the index number of its pointer in
 * the index file, to be deleted, by writing a special character as the
 * field delimiter indicator.
 *
 * For usage, delstr -h
 * version 0.2.0, first pass, September 25, 1996
 * 0.4.2, April 12, 1997.  Revision for libstr.
 * 0.5.0, June 17, 1997.  Minor revisions.
 *
 * Copyright (c) 1996 Amelia A. Lewis
 * Distributed under GPL.  See COPYING.
 *
 * $Id: delstr.c 13 2004-03-27 20:21:34Z amyzing $
 */

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include "../lib/strdb.h"
#include "../lib/maintdb.h"
#include "../lib/errdb.h"

char index_fn[MAXPATHLEN],
     *progname;
FILE *indexf, 
     *dataf;

unsigned long recnum[255];
unsigned char num_recs = 0;

int babble_flag = 0;
int gibberish_flag = 0;
int purge_flag = 0;

void
usage(char *calledas)
{
    sdb_version(calledas, NULL);
    fprintf(stderr, "Usage:\n\t%s [-fhv] file rnum [rnum ...]\n\n", calledas);
    fprintf(stderr, "\t-f\tForce use of outdated index file.\
\n	-h	Print this usage message.\
\n	-p	Purge (instead of marking deleted).\
\n	-v	Be verbose, report trivial statistics.\
\n	-V	Report version and exit.\
\nfile:		Name of the index file.\
\nrnum:           Index(es) of the pointer(s) to record(s) to delete.\n");
    exit(sdb_EUSAGE);
}

void
getargs(int ac, char *av[])
{
    int i;
    
    if (strrchr(av[0], '/'))
        progname = strdup(strrchr(av[0], '/'));
    else
        progname = strdup(av[0]);
    while ((i = getopt(ac, av, "fhvV")) != EOF)
        switch (i)
        {
            case 'f':
                gibberish_flag++;
                break;
            case 'p':
                purge_flag++;
                break;
            case 'v':
                babble_flag++;
                break;
            case 'V':
                sdb_version(progname, NULL);
                exit(0);
                break;
            case 'h':
            case ':':
            case '?':
            default:
                usage(av[0]);
                break;
        }
    
    av += optind;
    if (*av)
    {
       strcpy(index_fn, *av);
    }
    else 
        usage(progname);
    if (*(av + 1))
    {
        while (*++av)
        {
            recnum[num_recs] = strtoul(*av, NULL, 10);
            num_recs++;
        }
    }
    else 
        usage(progname);
}

void 
getfiles(sdb_header *sdbhp)
{
    int errcod = 0, i;
    
    indexf = sdb_findindex(index_fn, &errcod, gibberish_flag, purge_flag);
    if (!indexf)
    {
        fprintf(stderr, "%s: %s", progname, sdb_err_string[errcod]);
        exit(errcod);
    }
    
    for (i = 0; i < num_recs; i++)
    {
        if (recnum[i] > sdbhp->sdb_total)
        {
            fprintf(stderr, "%s: %lu: %s (%lu available)\n", progname, recnum[i], sdb_err_string[sdb_EIDXRANGE], sdbhp->sdb_total);
            exit(1);
        }
    }

    dataf = fopen(sdbhp->sdb_filename, "r+");
    if (!(dataf))
    {
        perror(sdbhp->sdb_filename);
        exit(sdb_ENOENT);
    }
}

int 
main(int argc, char *argv[])
{
    sdb_header sdbh;
    sdb_recptr *srp = NULL;
    int i;
    
    getargs(argc, argv);
    getfiles(&sdbh);
    for (i = 0; i < num_recs; i++)
    {
        srp = sdb_pointer(indexf, recnum[i]);
        if (!srp)
        {
            fprintf(stderr, "Record not found.\n");
            exit(sdb_ECHAOS);
        }
        else
        {
            if (!purge_flag)
                sdb_deleterec(dataf, srp->data_os, srp->data_size);
            else
                sdb_purgerec(indexf, recnum[i]);
        }
    }

    fclose(dataf);
    fclose(indexf);
    return 0;
}
