/*
 * dumpstr.c
 *
 * Part of the strings db toolkit.
 *
 * dumpstr accepts an indexfile as input, and outputs in format readable by
 * idxstr (or a similar utility, presumably).  The combination of idxstr
 * then dumpstr can leave a file in sorted order, normalize the file (empty
 * records and such are removed), and so on.  It doesn't retrieve subsets,
 * though; use grepstr.
 *
 * For usage, try dumpstr -h
 *
 * Yes, of *course* a toolkit full of "sters" has to have a dumpster!  Maybe
 * a mister, a sister, a monster ....
 *
 * version 0.2.0, apparently working.  September 22, 1996.
 * version 0.2.1 update, with findindex.  December 4, 1996.
 *    Made second file name optional, with default output to stdout.
 * 0.4.1, April 11, 1997.  Minor revision for "library" updates.
 * 0.5.0, June 17, 1997.  More minor revisions.
 *
 * Copyright (c) 1996 Amelia A. Lewis
 * Distributed under GPL.  See COPYING.
 *
 * $Id: dumpstr.c 13 2004-03-27 20:21:34Z amyzing $
 */

#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <ctype.h>
#include "../lib/strdb.h"
#include "../lib/errdb.h"

char index_fn[MAXPATHLEN], 
     oput_fn[MAXPATHLEN],
     *progname;
FILE *indexf, 
     *dataf, 
     *oputf;

int babble_flag = 0;
int gibberish_flag = 0;

char new_sc = '\0';

void 
usage(char *calledas)
{
    sdb_version(calledas, NULL);
    fprintf(stderr, "Usage:\n\t%s [-FhvV] [-c sch] indexfile [outputfile]\n\n", calledas);
    fprintf(stderr, "\t-c sch\tReplace the current signal character with 'sch'.\
\n	-F	Force output using outdated pointers (may lose data).\
\n	-h	Print this usage message.\
\n	-v	Be verbose, babble meaningless statistics.\
\n	-V	Report version and exit.\
\nindexfile:	Name of the index file to use.\
\noutputfile:	Name of the file to write data to (default: stdout)\n");
    exit(1);
}

void 
getargs(int ac, char *av[])
{
    int i;
   
    if (strrchr(av[0], '/'))
        progname = strdup(strrchr(av[0], '/'));
    else
        progname = strdup(av[0]);
    while ((i = getopt(ac, av, "c:FhvV")) != EOF)
        switch (i)
        {
            case 'c':
                if (isgraph(*optarg))
                    new_sc = *optarg;
                break;
            case 'F':
                gibberish_flag++;
                break;
            case 'v':
                babble_flag++;
                break;
            case 'V':
                sdb_version(progname, NULL);
                exit(0);
                break;
            case 'h':
            case ':':
            case '?':
            default:
                usage(progname);
                break;
        }
   
    av += optind;
    if (*av)
    {
        strcpy(index_fn, *av);
    }
    else 
        usage(progname);
    if (*++av)
    {
        strcpy(oput_fn, *av);
    }
    else
        strcpy(oput_fn, "-");
}

void 
getfiles(sdb_header *sdbhp)
{
    int fec = 0;
    
    indexf = sdb_findindex(index_fn, &fec, gibberish_flag, 0);
    if (!indexf)
    {
        fprintf(stderr, "%s: %s", progname, sdb_err_string[fec]);
        exit(fec);
    }
    rewind(indexf);
    fread((sdb_header *)&sdbhp, sizeof(sdb_header), 1, indexf);

    dataf = fopen(sdbhp->sdb_filename, "r");
    if (!(dataf))
    {
        perror(sdbhp->sdb_filename);
        exit(sdb_ENOENT);
    }

    if (strcmp(oput_fn, "-") == 0) oputf = stdout;
    else oputf = fopen(oput_fn, "w+");
    if (!(oputf))
    {
        perror(oput_fn);
        exit(sdb_ENOENT);
    }
}

int 
main(int argc, char *argv[])
{
    sdb_header sdbh;
    int nptrs;
    sdb_recptr *srp = NULL;
    char *fptr;
    
    getargs(argc, argv);
    getfiles(&sdbh);

    for (nptrs = 0 ; nptrs < sdbh.sdb_total ; nptrs++)
    {
        srp = sdb_nextptr(indexf);
        if (srp)
        {
            if (sdb_recalloc(dataf, srp) > 0)
            {
                if (new_sc) do
                {
                    fptr = sdb_ndinr(srp->data, sdbh.sdb_sigch);
                    if (fptr[1] == sdbh.sdb_sigch) fptr[1] = new_sc;
                } while (*fptr != '\0');
                fprintf(oputf, "%s", srp->data);
                sdb_recfree(srp);
            }
            free(srp);
            srp = NULL;
        }
    }

    fclose(indexf);
    fclose(dataf);
    fclose(oputf);
    if (babble_flag)
    {
        fprintf(stderr, "%d records dumped.\n", nptrs);
        if (new_sc) fprintf(stderr, "Signal character changed to: %c\n", new_sc);
        if (gibberish_flag) fprintf(stderr, "Some records may be incomplete.\n");
        fprintf(stderr, "Output to: %s\n", oput_fn);
    }
    return 0;
}
