/*
 * keystr.c
 * part of the strings db package
 *
 * keystr creates a file containing keywords from a specified field, with
 * associated index numbers of pointers in the index file.
 *
 * For usage instructions, use keystr -h
 * 
 * 0.4.0, March 24 1997, first pass at a working version.
 * 6 April 1997: probably made it find repeated fields, who knows?  Seems
 *    to have a glitch, putting ^Q at the end of some keys.  (fixed.
 *    fencepost.  *sigh*)
 * 0.4.3, June 17, 1997: added minimum token length option
 *
 * Copyright (c) 1997 Amelia A. Lewis
 * Distributed under GPL.  See COPYING.
 *
 * $Id: keystr.c 13 2004-03-27 20:21:34Z amyzing $
 */

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <sys/stat.h>
#include "../lib/strdb.h"
#include "../lib/keydb.h"
#include "../lib/errdb.h"

FILE *	keyf,
     *	indexf,
     *	dataf,
     *  stopf;

char	key_fn[MAXPATHLEN],
	index_fn[MAXPATHLEN],
	data_fn[MAXPATHLEN],
	stop_fn[MAXPATHLEN],
	*progname;

unsigned char stop_threshold = 50;
unsigned char min_length = 3;
char keyfield = '\0';
char signal_char;

int tkeys = 0;
int babble_flag = 0;
int gibberish_flag = 0;
int update_flag = 0;

typedef struct
{
    void *next;
    sdb_key k;
} key_list;

void
usage(char *calledas)
{
    sdb_version(calledas, NULL);
    fprintf(stderr, "Usage\n\t%s [-Fhuv] [-f fld] [-m min] [-t num] indexfile [.ext|keyfile]\
\n\
\n	-f fld	Create keys based on contents of field 'fld'.\
\n	-F	Force use of an outdated index file (gibberish).\
\n	-h	Print this usage message.\
\n	-m min	Minimum length of a key, in characters (default 3).\
\n	-s file	Exclude stopwords appearing in \"file\".\
\n	-t num	Exclude keys appearing in num%% of records (default 50%%).\
\n	-u	Use key field and threshold in previous version (update).\
\n	-v	Be verbose.\
\n	-V	Report version and exit.\
\n\
\nindexfile:	Name of the index file (created by idxstr or equivalent).\
\nkeyfile:	Name to give created key file.\
\n.ext:		Extension to give created key file, with same base as index.\
\n", calledas);
    exit(sdb_EUSAGE);
}

void
getargs(int ac, char *av[])
{
    int c;

    stop_fn[0] = '\0';
    if (strrchr(av[0], '/'))
        progname = strdup(strrchr(av[0], '/'));
    else
        progname = strdup(av[0]);
    while ((c = getopt(ac, av, "Ff:hm:st:uvV")) != EOF)
         switch (c)
         {
             case 'F':
                 gibberish_flag++;
                 break;
             case 'f':
                 if (isgraph(*optarg))
                     keyfield = *optarg;
                 break;
             case 'm':
                 if (*optarg)
                     min_length = (unsigned char)strtoul(optarg, NULL, 10);
                 break;
             case 's':
                 if (*optarg)
                     strcpy(stop_fn, optarg);
                 break;
             case 't':
                 if (*optarg)
                     stop_threshold = (unsigned char)strtoul(optarg, NULL, 10);
                 break;
             case 'u':
                 update_flag++;
                 break;
             case 'v':
                 babble_flag++;
                 break;
             case 'V':
                 sdb_version(progname, NULL);
                 exit(0);
                 break;
             case 'h':
             case ':':
             case '?':
             default:
                 usage(progname);
                 break;
         }

     av += optind;
     if (*av)
         strcpy(index_fn, *av);
     else
         usage(progname);
    if (index_fn[0] == '-')
    {
        fprintf(stderr, "Input from stdin not allowed.\n");
        usage(progname);
    }
    if (*++av)
        strcpy(key_fn, *av);
    else
        key_fn[0] = '\0';
}

void
getfiles(void)
{
    sdb_header sdbh;
    sdb_key_head sk;
    int filerr;
    char exten[5], *iex;
    
    /* get the index file, check for magic */
    indexf = sdb_findindex(index_fn, &filerr, gibberish_flag, 0);
    if (!indexf)
    {
        fprintf(stderr, "Error: %s: %s", index_fn, sdb_err_string[filerr]);
        exit(filerr);
    }
    rewind(indexf);
    fread((sdb_header *)&sdbh, sizeof(sdb_header), 1, indexf);
    signal_char = sdbh.sdb_sigch;

    /* set key field and key file name, if not already done */
    if (!keyfield)
        keyfield = sdbh.sdb_sorton ? sdbh.sdb_sorton : sdbh.sdb_sigch;
    if (!key_fn[0])
    {
        strcpy(key_fn, index_fn);
        iex = strrchr(key_fn, '.');
        if (iex)
            *iex = '\0';
        if (keyfield == signal_char)
            strcat(key_fn, ".key");
        else
        {
            sprintf(exten, ".%cky", keyfield);
            strcat(key_fn, exten);
        }
    }

    strcpy(data_fn, sdbh.sdb_filename);
    dataf = fopen(data_fn, "r");
    if (!dataf)
    {
        perror(data_fn);
        exit(sdb_ENOENT);
    }

    /* if updating, open the key file, read flags, set them */
    if (update_flag)
    {
        keyf = fopen(key_fn, "r");
        if (keyf)
        {
            rewind(keyf);
            if (fread((sdb_key_head *)&sk, sizeof(sdb_key_head), 1, keyf))
            {
                keyfield = sk.sdb_field;
                stop_threshold = sk.sdb_discthresh;
                min_length = sk.sdb_minlen;
            }
            fclose(keyf);
        }
        else if (babble_flag)
            fprintf(stderr, "Original file not found for update; using defaults\n");
    }
    /* create the keyfile */
    keyf = fopen(key_fn, "w+");
    if (!keyf)
    {
        perror(key_fn);
        exit(sdb_ENOENT);
    }
    
    /* If stop_fn isn't null, try to find the file of stopwords */
    if (*stop_fn)
    {
    	fprintf(stderr, "Stopword file not implemented.  Shoot the programmer.\n");
    }
}

void
init_header(sdb_key_head *sdbkhp)
{
    struct stat datastat;
    char *tempfn;

    sdbkhp->sdb_major = sdb_VMAJOR;
    sdbkhp->sdb_minor = sdb_VMINOR;
    sdbkhp->sdb_patch = sdb_VPATCH;
    sdbkhp->sdb_field = keyfield;
    strcpy(sdbkhp->sdb_magic, sdb_MAGIC);
    sdbkhp->sdb_total = 0;
    sdbkhp->sdb_discthresh = stop_threshold;
    sdbkhp->sdb_minlen = min_length;
    stat(index_fn, &datastat);
    sdbkhp->sdb_mtime = datastat.st_mtime;
    tempfn = realpath(index_fn, sdbkhp->sdb_indexfile);
    if (!tempfn)
        strcpy(sdbkhp->sdb_indexfile, index_fn);
}

key_list *
new_keylist(unsigned short w, char *k)
{
    key_list *nkl = NULL;

    nkl = (key_list *)malloc(sizeof(key_list));
    nkl->next = NULL;
    nkl->k.key = strdup(k);
    nkl->k.nindices = 1;
    nkl->k.indices = (unsigned short *)malloc(sizeof(unsigned short));
    nkl->k.indices[0] = w;
    return nkl;
}

char *
normalize_field(char *f)
{
    char *fend, *fnext;
    
    if (!f)
        return NULL;
    fend = sdb_ndinr(f, signal_char);
    *fend = '\0';
    if (*(fend + 1))
        fnext = fend + 1;
    else
        fnext = NULL;
    fend = f;
    while (*fend)
    {
        *fend = (char)tolower(*fend);
        fend++;
    }
    return fnext;
}

key_list *
rec_tokenize(key_list *kl, char *r, unsigned short which)
{
    char *fstart, *toke;
    char delimiters[] = "\n\t ~!@#$%^&*()_+-={}[]:\";'<>?,./|\\";
    key_list *tkl = NULL, *pkl = NULL;
    int found;

    if (!r)
        return kl;
    fstart = r;
    
    /* repeatedly get a token from the field */
    if ((toke = strtok(fstart, delimiters)) == NULL)
        return kl;
    else do
    {
        if (!kl && (strlen(toke) >= min_length))		/* first */
        {
            kl = new_keylist(which, toke);
            if (babble_flag)
                fprintf(stderr, "Adding token: %s\n", toke);
        }
        /* check to see if this token is already in the list */
        else if (strlen(toke) >= min_length)
        {
            tkl = kl;
            found = 0;
            while (tkl)
            {
                pkl = tkl;
                if (!strcmp(tkl->k.key, toke))
                {
                    if (which != tkl->k.indices[tkl->k.nindices - 1])
                    {
                        tkl->k.nindices++;
                        /* this is grody.  check for errors, too */
                        tkl->k.indices = (unsigned short *)realloc(tkl->k.indices, sizeof(unsigned short) * tkl->k.nindices);
                        tkl->k.indices[tkl->k.nindices - 1] = which;
                    }
                    found++;
                    break;
                }
                tkl = (key_list *)tkl->next;
            }
            if (!found)
            {
                if (babble_flag)
                    fprintf(stderr, "Adding token: %s\n", toke);
                tkl = new_keylist(which, toke);
                pkl->next = (key_list *)tkl;
            }
        }
    } while ((toke = strtok(NULL, delimiters)) != NULL);
    return kl;
}

key_list *
weed_list(key_list *kl, unsigned long trecs)
{
    key_list *thiskl, *deletekl = NULL, *startkl = NULL, *lastkl;

    if (!kl)
        return NULL;
    thiskl = kl;
    while (!startkl && thiskl)
    {
        if ((thiskl->k.nindices * 100 / trecs) >= stop_threshold)
            deletekl = thiskl;
        else
            startkl = thiskl;
        thiskl = (key_list *)thiskl->next;
        if (deletekl)
        {
            if (babble_flag)
                fprintf(stderr, "Deleting key: %s\n", deletekl->k.key);
            sdb_key_free(&(deletekl->k));
            free(deletekl);
            deletekl = NULL;
        }
    }
    if (!thiskl)
        return NULL;
    /* now we have a start-of-list */
    lastkl = startkl;
    while (thiskl)
    {
        if((thiskl->k.nindices * 100 / trecs) >= stop_threshold)
            deletekl = thiskl;
        thiskl = (key_list *)thiskl->next;
        if (deletekl)
        {
            if (babble_flag)
                fprintf(stderr, "Deleting key: %s\n", deletekl->k.key);
            lastkl->next = thiskl;	/* even if NULL */
            sdb_key_free(&(deletekl->k));
            free(deletekl);
            deletekl = NULL;
        }
        else
            lastkl = (key_list *)lastkl->next;
    }
    return startkl;
}

int
main (int argc, char *argv[])
{
    sdb_key_head sdbkh;
    sdb_header sdbh;
    sdb_recptr *curptr = NULL;
    key_list *keylist = NULL,	/* start of list */
             *pkey = NULL;
    unsigned short i;
    char *field, *nextf;
    char cmdstr[MAXPATHLEN + 13];

    getargs(argc, argv);
    getfiles();
    init_header(&sdbkh);
    rewind(indexf);
    fread((sdb_header *)&sdbh, sizeof(sdb_header), 1, indexf);
    fseek(indexf, sizeof(sdb_header), SEEK_SET);
    for (i = 1 ; i <= sdbh.sdb_total ; i++)
    {
        curptr = sdb_nextptr(indexf);
        if (curptr->data_size == sdb_recalloc(dataf, curptr))
        {
            if (signal_char == keyfield)
            {
                nextf = normalize_field(curptr->data);
                keylist = rec_tokenize(keylist, curptr->data, i);
            }
            else
            {
                field = sdb_finr(curptr->data, signal_char, keyfield);
                while (field)
                {
                    nextf = normalize_field(field);
                    keylist = rec_tokenize(keylist, field, i);
                    field = sdb_finr(nextf, signal_char, keyfield);
                }
            }
        }
        if (curptr)
        {
            sdb_recfree(curptr);
            free(curptr);
        }
    }

    fclose(indexf);
    fclose(dataf);

    if (stop_threshold < 100)
        keylist = weed_list(keylist, sdbh.sdb_total);
    pkey = keylist;
    while (pkey)
    {
        tkeys++;
        pkey = (key_list *)pkey->next;
    }
    sdbkh.sdb_total = tkeys;
    if (babble_flag)
        fprintf(stderr, "Writing %u keys: ", tkeys);
    fwrite((sdb_key_head *)&sdbkh, sizeof(sdb_key_head), 1, keyf);
    pkey = keylist;
    while (pkey)
    {
        if (babble_flag)
            fprintf(stderr, ".");
        sdb_key_write(&pkey->k, keyf);
        pkey = (key_list *)pkey->next;
    }
    if (babble_flag)
        fprintf(stderr, " done.\n");
    
    /* clean up keylist */
    
    fclose(keyf);
    if (babble_flag)
    {
    	sprintf(cmdstr, "lsstr -Akv %s", key_fn);
    	system(cmdstr);
    }
    return(0);
}
