/*
 * lsstr.c
 *
 * Part of the strings db toolkit.
 *
 * lsstr reads the information in an index file and displays it on stdout,
 * in more or less comprehensible fashion (or so we fondly hope).
 *
 * version 0.2.0, seems solid, September 22, 1996.
 * version 0.2.1, a bit later.
 * November 26, 1996: started revision, to permit multiple parameters
 * with simple output, making old version into "verbose" (i.e., human-readable)
 * Switched to using the findindex function.
 * 0.4.0.  Completed 6 April 1997, complete revision of structure and
 *    semantics.  Reads key files.  Default output -nv.  Output without
 *    -v is intended for feeding to other utilities.
 * 
 * Copyright (c) 1996 Amelia A. Lewis
 * Distributed under GPL.  See COPYING.
 *
 * $Id: lsstr.c 13 2004-03-27 20:21:34Z amyzing $
 */

#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <time.h>
#include "../lib/strdb.h"
#include "../lib/keydb.h"
#include "../lib/errdb.h"

char fn[MAXPATHLEN];
char *progname;
FILE *f;

int babble_flag = 0;
int app_flag = 0;
int file_flag = 0;
int num_flag = 0;
int sort_flag = 0;
int sortflags_flag = 0;
int version_flag = 0;
int signal_flag = 0;
int date_flag = 0;
int Date_flag = 0;
int key_flag = 0;
int thresh_flag = 0;
int magic_flag = 0;
int pointer_flag = 0;
int set_flag = 0;
int length_flag = 0;

void usage(char *calledas)
{
    sdb_version(calledas, NULL);
    fprintf(stderr, "Usage:\n\t%s [-aAcdDfhkmnpsStvV] filename\n\n", calledas);
    fprintf(stderr, "\t-a\tOutput application-specific flags\
\n	-A	Output everything relevant\
\n	-c	Output signal character\
\n	-d	Output modification time stored in index/key file\
\n	-D	Output modification time of indexed/keyed file\
\n	-f	Output name of indexed/keyed file\
\n	-h	Print this usage message\
\n	-k	Examine key file instead of index\
\n	-l	Output minimum key length\
\n	-m	Output magic\
\n	-n	Output number of pointers/keys\
\n	-p	Dump the pointers\
\n	-q	Don't be verbose (suppress default)\
\n	-s	Output identifier of sort field (key files: key field)\
\n	-S	Output sort flags\
\n	-t	Output stop threshold (key files only)\
\n	-v	Be verbose (print human-readable)\
\n	-V	Print version of strdb used to create index/key file\
\nfilename:	strdb index or key file to examine\
\n\
\nIf called with no command line parameters, default output is -nv\
\nMultiple options are output in the order defined in the header.\n");
    exit(sdb_EUSAGE);
}

void getargs(int ac, char *av[])
{
    int i;
    
    if (strrchr(av[0], '/'))
        progname = strdup(strrchr(av[0], '/'));
    else
        progname = strdup(av[0]);
    while ((i = getopt(ac, av, "aAcdDfhklmnpqsStvV")) != EOF)
        switch(i)
        {
            case 'a':
                app_flag++;
                set_flag++;
                break;
            case 'A':
                app_flag++;
                signal_flag++;
                date_flag++;
                Date_flag++;
                file_flag++;
                num_flag++;
                sort_flag++;
                sortflags_flag++;
                thresh_flag++;
                version_flag++;
                set_flag++;
                length_flag++;
                break;
            case 'c':
                signal_flag++;
                set_flag++;
                break;
            case 'd':
                date_flag++;
                set_flag++;
                break;
            case 'D':
                Date_flag++;
                set_flag++;
                break;
            case 'f':
                file_flag++;
                set_flag++;
                break;
            case 'k':
                key_flag++;
                break;
            case 'l':
                length_flag++;
                set_flag++;
                break;
            case 'm':
                magic_flag++;
                set_flag++;
                break;
            case 'n':
                num_flag++;
                set_flag++;
                break;
            case 'p':
                pointer_flag++;
                break;
            case 'q':
                babble_flag = 0;
                set_flag++;
                break;
            case 's':
                sort_flag++;
                set_flag++;
                break;
            case 'S':
                sortflags_flag++;
                set_flag++;
                break;
            case 't':
                thresh_flag++;
                set_flag++;
                break;
            case 'v':
                babble_flag++;
                break;
            case 'V':
                version_flag++;
                set_flag++;
                break;
            case 'h':
            case ':':
            case '?':
            default:
                usage(progname);
                break;
        }
    
    if (!set_flag)
    {
        babble_flag++;
        num_flag++;
    }
    av += optind;
    if (*av)
    {
        strcpy(fn, *av);
    }
    else usage(progname);
}

void
getfiles()
{
    int filerr;

    if (key_flag)
    {
        f = fopen(fn, "r");
        if (!f)
        {
            perror(fn);
            exit(sdb_ENOENT);
        }
        filerr = sdb_unicorn(f, 1);
        if (filerr)
        {
            if (babble_flag)
            {
                fprintf(stderr, "%s: %s", fn, sdb_err_string[filerr]);
                if (filerr != sdb_EIDXDATE)
                    exit(filerr);
            }
        }
    }
    else
    {
        f = sdb_findindex(fn, &filerr, 1, 0);	/* gibberish ON */
        if (!f)
        {
            if (babble_flag) 
            {
                fprintf(stderr, "%s: %s", fn, sdb_err_string[filerr]);
                if (filerr != sdb_EIDXDATE)
                    exit(filerr);
             }
        }
    }
}

void
display_index(sdb_header *psdbh)
{
    struct stat datastat;
    unsigned long iterator;
    sdb_recptr *thisp;
    int i;

    if (version_flag)
    {
        if (babble_flag)
        {
            printf("strdb version %d.%d.%d\n", psdbh->sdb_major, psdbh->sdb_minor, psdbh->sdb_patch);
        }
        else
            printf("%d.%d.%d ", psdbh->sdb_major, psdbh->sdb_minor, psdbh->sdb_patch);
    }
    if (signal_flag)
    {
        if (babble_flag)
        {
            printf("Signal character: %c\n", psdbh->sdb_sigch);
        }
        else
            printf("%c ", psdbh->sdb_sigch);
    }
    if (sort_flag)
    {
        if (babble_flag)
        {
            if (!psdbh->sdb_sorton)
                printf("Not sorted\n");
            else
            {
                if (psdbh->sdb_sorton == psdbh->sdb_sigch)
                    printf("Sorted on the ambiguous field");
                else
                    printf("Sorted on field %c", psdbh->sdb_sorton);
            }
        }
        else
            psdbh->sdb_sorton ? printf("%c ", psdbh->sdb_sorton)
                              : printf("NULL ");
    }
    if (sortflags_flag)
    {
        if (babble_flag)
        {
            if (psdbh->sdb_sorton)
            {
                if (psdbh->sdb_sortflags & sdb_SORT_CASE)
                    printf(", preserving case");
                else
                    printf(", ignoring case");
                if (psdbh->sdb_sortflags & sdb_SORT_REVERSE)
                    printf(", reverse order");
                if (psdbh->sdb_sortflags & sdb_SORT_IWHITE)
                    printf(", ignoring whitespace");
                if (psdbh->sdb_sortflags & sdb_SORT_ALNUM)
                    printf(", alphanumeric only");
                if (psdbh->sdb_sortflags & sdb_SORT_NUMBER)
                    printf(", as integer");
                printf("\n");
            }
        }
        else
           printf("%u ", psdbh->sdb_sortflags);
    }
    if (app_flag)
    {
        if (babble_flag)
        {
            if (psdbh->sdb_appflags)
                printf("Application-specific flags: ");
            for (i = 0; i < 7; i++)
                if (psdbh->sdb_appflags & (1 << i))
                    printf("%d ", i);
            if (psdbh->sdb_appflags)
                printf("\n");
        }
        else
            printf("%u ", psdbh->sdb_appflags);
    }
    if (magic_flag)
    {
        if (babble_flag)
        {
            printf("Magic: %s\n", psdbh->sdb_magic);
        }
        else
            printf("%s ", psdbh->sdb_magic);
    }
    if (num_flag)
    {
        if (babble_flag)
        {
            printf("Total records: %lu\n", psdbh->sdb_total);
        }
        else
            printf("%lu ", psdbh->sdb_total);
    }
    if (date_flag)
    {
        if (babble_flag)
        {
            printf("Stored data file modification time: %s", ctime(&psdbh->sdb_mtime));
        }
        else
            printf("%lu ", psdbh->sdb_mtime);
    }
    if (Date_flag)
    {
        if (stat(psdbh->sdb_filename, &datastat) == -1)
            datastat.st_mtime = 0;
        if (babble_flag)
        {
            printf("Current data file modification time: %s", ctime(&datastat.st_mtime));
        }
        else
            printf("%lu ", datastat.st_mtime);
    }
    if (file_flag)
    {
        if (babble_flag)
        {
            printf("Data file indexed: %s\n", psdbh->sdb_filename);
        }
        else
            printf("%s ", psdbh->sdb_filename);
    }
    if (pointer_flag)
    {
        for (iterator = 1; iterator <= psdbh->sdb_total; iterator++)
        {
            thisp = sdb_pointer(f, iterator);
            if (babble_flag)
            {
                printf("Pointer %lu: offset %lu, size %u\n", iterator, thisp->data_os, thisp->data_size);
            }
            else
                printf("%lu:%lu:%u ", iterator, thisp->data_os, thisp->data_size);
        }
    }
    if (!babble_flag)
        printf("\n");
}

void
display_key(sdb_key_head *psdbkh)
{
    struct stat datastat;
    unsigned short iterator;
    sdb_key thisk;
    int i;

    if (version_flag)
    {
        if (babble_flag)
        {
            printf("strdb version %d.%d.%d\n", psdbkh->sdb_major, psdbkh->sdb_minor, psdbkh->sdb_patch);
        }
        else
            printf("%d.%d.%d ", psdbkh->sdb_major, psdbkh->sdb_minor, psdbkh->sdb_patch);
    }
    if (sort_flag)
    {
        if (babble_flag)
        {
            printf("Field keyed: %c\n", psdbkh->sdb_field);
        }
        else
            printf("%c ", psdbkh->sdb_field);
    }
    if (magic_flag)
    {
        if (babble_flag)
        {
            printf("Magic: %s\n", psdbkh->sdb_magic);
        }
        else
            printf("%s ", psdbkh->sdb_magic);
    }
    if (num_flag)
    {
        if (babble_flag)
        {
            printf("Total keys: %u\n", psdbkh->sdb_total);
        }
        else
            printf("%u ", psdbkh->sdb_total);
    }
    if (thresh_flag)
    {
        if (babble_flag)
        {
            printf("Stop threshold: %u%%\n", psdbkh->sdb_discthresh);
        }
        else
            printf("%u ", psdbkh->sdb_discthresh);
    }
    if (length_flag)
    {
        if (babble_flag)
        {
            printf("Minimum key length: %u characters\n", psdbkh->sdb_minlen);
        }
        else
            printf("%u ", psdbkh->sdb_minlen);
    }
    if (date_flag)
    {
        if (babble_flag)
        {
            printf("Stored index file modification time: %s", ctime(&psdbkh->sdb_mtime));
        }
        else
            printf("%lu ", psdbkh->sdb_mtime);
    }
    if (Date_flag)
    {
        if (stat(psdbkh->sdb_indexfile, &datastat) == -1)
            datastat.st_mtime = 0;
        if (babble_flag)
        {
            printf("Current index file modification time: %s", ctime(&datastat.st_mtime));
        }
        else
            printf("%lu ", datastat.st_mtime);
    }
    if (file_flag)
    {
        if (babble_flag)
        {
            printf("Keyed using index file %s\n", psdbkh->sdb_indexfile);
        }
        else
            printf("%s ", psdbkh->sdb_indexfile);
    }
    if (pointer_flag)
    {
        /* iterate over the list of pointers.  print key:index,index,index[space] */
        for (iterator = 1; iterator <= psdbkh->sdb_total; iterator++)
        {
            sdb_key_read(&thisk, f);
            if (babble_flag)
            {
                printf("%s (%u indices): %u", thisk.key, thisk.nindices, thisk.indices[0]);
                for (i = 1; i < thisk.nindices; i++)
                    printf(", %u", thisk.indices[i]);
                printf("\n");
            }
            else
            {
                printf("%s", thisk.key);
                for (i = 0; i < thisk.nindices; i++)
                    printf(":%u", thisk.indices[i]);
                printf(" ");
            }
            sdb_key_free(&thisk);
        }
    }
    if (!babble_flag)
        printf("\n");
}

int main(int argc, char *argv[])
{
    sdb_header sdbh;
    sdb_key_head sdbkh;
    struct stat datastat;
    
    getargs(argc, argv);
    
    getfiles();
    
    rewind(f);
    if (key_flag)
    {
        fread((sdb_key_head *)&sdbkh, sizeof(sdb_key_head), 1, f);
        /* stat ? */
        display_key(&sdbkh);
    }
    else
    {
        fread((sdb_header *)&sdbh, sizeof(sdb_header), 1, f);
        if (stat(sdbh.sdb_filename, &datastat) == -1)
        /* do something else */
            fprintf(stderr, "Error: %s, indexed by %s: %s", sdbh.sdb_filename, fn, sdb_err_string[sdb_ENOENT]);
        display_index(&sdbh);
    }

    fclose(f);
    return 0;
}
